import sys
import numpy as np
import pandas as pd
from fisher import pvalue
from scipy.stats import chi2_contingency
from statsmodels.stats.multitest import fdrcorrection

from unpast.utils.method import zscore
from lifelines import CoxPHFitter
from lifelines.statistics import logrank_test

def generate_exprs(data_sizes, g_size=5, frac_samples=[0.05, 0.1, 0.25, 0.5], m=2.0, std=1,
                   z=True,
                   outdir="./", outfile_basename="",
                   g_overlap=False, s_overlap=True,
                   seed=42, add_coexpressed=[]):
    n_genes, N = data_sizes
    biclusters = {}
    coexpressed_modules = []

    # generate background model
    np.random.seed(seed)
    exprs = pd.DataFrame(np.random.normal(loc=0, scale=1.0, size=(n_genes, N)))
    exprs.columns = ["s_" + str(x) for x in exprs.columns.values]
    exprs.index = ["g_" + str(x) for x in exprs.index.values]

    # implant bicluster
    N = exprs.shape[1]
    bic_g = []
    bic_s = []
    bg_g = set(exprs.index.values).difference(set(bic_g))
    bg_s = set(exprs.columns.values).difference(set(bic_s))
    bicluster_genes = []
    for s_frac in frac_samples:
        s_size = int(s_frac * N)
        # select random sets of samples and genes from the background
        bic_genes = list(np.random.choice(list(bg_g), size=g_size, replace=False))
        bic_samples = list(np.random.choice(list(bg_s), size=s_size, replace=False))
        bic_g += bic_genes
        bic_s += bic_samples
        # identify samples outside the bicluster
        if not g_overlap:
            bg_g = bg_g.difference(set(bic_g))
        if not s_overlap:
            bg_s = bg_s.difference(set(bic_s))
        # generate bicluster
        biclusters[s_frac] = {"genes": set(bic_genes), "samples": set(bic_samples), "frac": s_frac,
                              "n_genes": len(bic_genes), "n_samples:": len(bic_samples)}
        bic_exprs = np.random.normal(loc=m, scale=std, size=(g_size, s_size))
        # implant biclusters
        exprs.loc[bic_genes, bic_samples] += bic_exprs
        bicluster_genes += bic_genes

    # add modules of co-expressed genes 
    bg_g = set(exprs.index.values).difference(set(bicluster_genes))
    r = 0.5
    for module in add_coexpressed:
        module_genes = list(np.random.choice(list(bg_g), size=module, replace=False))
        n = exprs.loc[module_genes[0], :]
        for i in range(1, module):
            n_i = n * r + np.sqrt(1 - r ** 2) * exprs.loc[module_genes[i], :]
            exprs.loc[module_genes[i], :] = n_i
        print("\tco-exprs. module ", module, "r=",
              (exprs.loc[module_genes, :].T.corr().sum().sum() - module) / (module ** 2 / 2 - module))
        coexpressed_modules.append(module_genes)

    if z:
        # center to 0 and scale std to 1
        exprs = zscore(exprs)
    biclusters = pd.DataFrame.from_dict(biclusters).T
    # biclusters.set_index("frac",inplace = True,drop=True)
    biclusters_ = biclusters.copy()

    if outfile_basename:
        # overlap extension
        if s_overlap:
            if g_overlap:
                overlap_ext = ",overlap=yes"
            else:
                overlap_ext = ",overlap=s"
        elif g_overlap:
            overlap_ext = ",overlap=g"
        else:
            overlap_ext = ",overlap=no"
        # save expressions 
        exprs_file = outdir + "/" + outfile_basename + ".n_genes=" + str(g_size) + ",m=" + str(m) + ",std=" + str(std)
        exprs_file += overlap_ext + ".exprs_z.tsv"
        print("expressions:", exprs_file)
        exprs.to_csv(exprs_file, sep="\t")

        # save ground truth 
        biclusters["n_genes"] = biclusters["genes"].apply(lambda x: len(x))
        biclusters["n_samples"] = biclusters["samples"].apply(lambda x: len(x))
        biclusters["genes"] = biclusters["genes"].apply(lambda x: " ".join((map(str, sorted(x)))))
        biclusters["samples"] = biclusters["samples"].apply(lambda x: " ".join((map(str, sorted(x)))))

        biclusters_file = outdir + "/" + outfile_basename + ".n_genes=" + str(g_size) + ",m=" + str(m) + ",std=" + str(
            std)
        biclusters_file += overlap_ext + ".biclusters.tsv"
        biclusters.to_csv(biclusters_file, sep="\t")
        print("true bilusters:", biclusters_file)
        biclusters.to_csv(biclusters_file, sep="\t")

    return exprs, biclusters_, coexpressed_modules

def make_ref_groups(subtypes, annotation,exprs):
    import copy
    from collections import OrderedDict
    # prepared a dict of subtype classifications {"class1":{"subt1":[],"subt2":[]},"class2":{"subtA":[],"subtB":[]}}
    all_samples = set(exprs.columns.values).intersection(set(subtypes.index.values))
    s = sorted(all_samples)
    
    pam50 = make_known_groups(subtypes.loc[s,:], exprs.loc[:,s],target_col = "PAM50",verbose=False)
    lum = {}
    lum["Luminal"] = pam50["LumA"].union(pam50["LumB"]).intersection(all_samples)
    scmod2 = make_known_groups(subtypes.loc[s,:], exprs.loc[:,s], target_col = 'SCMOD2',verbose=False)
    claudin = {} 
    claudin["Claudin-low"] = set(subtypes.loc[subtypes['claudin_low']==1,:].index.values).intersection(all_samples)
    
    
    ihc = {}
    for x in ["IHC_HER2","IHC_ER","IHC_PR"]:
        ihc[x] = set(annotation.loc[annotation[x]=="Positive",:].index.values)#.intersection(all_samples)
    ihc["IHC_TNBC"] = set(annotation.loc[annotation["IHC_TNBC"]==1,:].index.values)#.intersection(all_samples)
    
    pam50_lum = copy.copy(pam50)
    del pam50_lum["LumA"]
    del pam50_lum["LumB"]
    pam50_lum["Luminal"] = lum["Luminal"]
    intrinsic = copy.copy(pam50_lum)
    intrinsic["Claudin-low"] = claudin["Claudin-low"]
    known_groups = OrderedDict({"PAM50":pam50_lum,"Intrinsic":intrinsic,"PAM50_AB":pam50,"SCMOD2":scmod2,"IHC":ihc})
    known_groups["Luminal"] = {"Luminal":pam50_lum["Luminal"]}
    known_groups["Basal"] = {"Basal":pam50["Basal"]}
    known_groups["Her2"] = {"Her2":pam50["Her2"]}
    known_groups["LumA"] = {"LumA":pam50["LumA"]}
    known_groups["LumB"] = {"LumB":pam50["LumB"]}
    known_groups["Normal"] = {"Normal":pam50["Normal"]}
    known_groups["Claudin-low"] = {"Claudin-low":claudin["Claudin-low"]}
    known_groups["IHC_HER2"] = {"IHC_HER2":ihc["IHC_HER2"]}
    known_groups["IHC_ER"] = {"IHC_ER":ihc["IHC_ER"]}
    known_groups["IHC_PR"] = {"IHC_PR":ihc["IHC_PR"]}
    known_groups["IHC_TNBC"] = {"IHC_TNBC":ihc["IHC_TNBC"]}
    known_groups["NET_kmeans"] = {"NET":set(subtypes.loc[subtypes["NET_km"]==1,:].index.values)}
    known_groups["NET_ward"] = {"NET":set(subtypes.loc[subtypes["NET_w"]==1,:].index.values)}
    return known_groups, all_samples

def make_known_groups(annot, exprs, target_col="genefu_z", verbose=False):
    samples = set(exprs.columns.values).intersection(set(annot.index.values))
    if verbose:
        print("Total samples:", len(samples), file=sys.stdout)
    annot = annot.loc[list(samples), :]
    groups = set(annot.loc[:, target_col].values)

    known_groups = {}
    for group in groups:
        if group == group:
            group_samples = set(annot.loc[annot[target_col] == group, :].index.values)
            group_samples = group_samples.intersection(samples)
            if len(group_samples) > int(len(samples) / 2):
                print("take complement of ", group, file=sys.stderr)
                group_samples = samples.difference(group_samples)
            known_groups[group] = group_samples  # {"set":group_samples,"complement": samples.difference(group_samples)}
            if verbose:
                print(group, round(len(group_samples) / len(samples), 2),
                      len(group_samples), len(samples.difference(group_samples)))
    return known_groups

def compare_gene_clusters(bics1,bics2, N):
    # N - total number of genes
    # finds best matched B1 -> B2 and B2 -> B1
    # calculates % of matched clusters, number of genes in matched clusters, 
    # and the average J index for best matches 
    bm = find_best_matching_biclusters(bics1,bics2,(N,0),by = "genes")
    bm = bm.dropna()
    bm2 = find_best_matching_biclusters(bics2,bics1,(N,0),by = "genes")
    bm2 = bm2.dropna()
    
    if "n_shared_genes" in bm.columns:
        bm = bm.loc[bm["n_shared_genes"]>1,:].sort_values(by="n_shared_genes",ascending = False)
    else:
        # no match -> remove all rows 
        bm = bm.head(0)
    if "n_shared_genes" in bm2.columns:
        bm2 = bm2.loc[bm2["n_shared_genes"]>1,:].sort_values(by="n_shared_genes",ascending = False)
    else:
        bm2 = bm.head(0)
    
    
    clust_similarity = {}
    # number of biclusters 
    clust_similarity["n_1"] = bics1.shape[0]
    clust_similarity["n_2"] = bics2.shape[0]
    #print("% matched biclusters:",bm.shape[0]/tcga_result.shape[0],bm2.shape[0]/metabric_result.shape[0])
    clust_similarity["percent_matched_1"] = bm.shape[0]/bics1.shape[0]
    clust_similarity["percent_matched_2"] = bm2.shape[0]/bics2.shape[0]
    
    #print("n matched genes:",bm.loc[:,"n_shared"].sum(),bm2.loc[:,"n_shared"].sum())
    if "n_shared_genes" in bm.columns:
        clust_similarity["n_shared_genes_1"] = bm.loc[:,"n_shared_genes"].sum()
        clust_similarity["avg_bm_J_1"] = bm.loc[:,"J"].mean()
    if "n_shared_genes" in bm2.columns:
        clust_similarity["n_shared_genes_2"] = bm2.loc[:,"n_shared_genes"].sum()
    #print("avg. J:",bm.loc[:,"J"].mean(),bm2.loc[:,"J"].mean())
        clust_similarity["avg_bm_J_2"] = bm2.loc[:,"J"].mean()
    
    
    return clust_similarity, bm, bm2

from sklearn.metrics import adjusted_rand_score

def calculate_perfromance(sample_clusters_, # data.Frame with "samples" column
                              known_groups, # dict={"classification1":{"group1":{"s1","s2",...},"group2":{...}, ...}}
                              all_samples, # set of all samples in input; needed for overlap p-value computations
                              performance_measure = "Jaccard",# must be "ARI" or "Jaccard"     
                              adjust_pvals = "B", # ["B", "BH", False] # correction for multiple testing
                              pval_cutoff = 0.05, # cutoff for p-values to select significant matches
                              min_SNR=0, min_n_genes=False,min_n_samples=1,
                              verbose=False):
    # select the sample set best matching the subtype based on p-value 
    # adj. overlap p-value should be:
    # below pval_cutoff, e.g. < 0.05
    # the lowest among overlap p-values computed for this sample set vs all subtypes
    # cluster with the highest J is chosen as the best match
    if sample_clusters_ is None or sample_clusters_.shape[0] == 0:
        return pd.DataFrame(),pd.DataFrame()
    
    if adjust_pvals not in ["B","BH", False]:
        print(adjust_pvals, "is not recognized, Bonferroni method will be used,",file = sys.stderr)
    sample_clusters = sample_clusters_[sample_clusters_["samples"].apply(lambda x: len(x)) >= min_n_samples]
    sample_clusters["n_samples"] = sample_clusters["samples"].apply(lambda x: len(x))
    if min_SNR:
        sample_clusters = sample_clusters[sample_clusters["SNR"] >= min_SNR]
    if min_n_genes:
        sample_clusters = sample_clusters[sample_clusters["genes"].apply(lambda x: len(x)) >= min_n_genes]
    
    if sample_clusters.shape[0] == 0:
        return pd.DataFrame(),pd.DataFrame()
    
    best_matches = []
    performances = {}
    for cl in known_groups.keys():
        # denominator for weights 
        N=0
        for subt in known_groups[cl].keys():
            N+= len(known_groups[cl][subt])
        if performance_measure == "ARI":
            pvals, is_enriched, performance = evaluate_overlaps_ARI(sample_clusters, known_groups[cl], all_samples)
        if performance_measure == "Jaccard":
            pvals, is_enriched, performance = evaluate_overlaps(sample_clusters, known_groups[cl], all_samples)
        if adjust_pvals:
            if adjust_pvals == "B":
                pvals  = pvals*pvals.shape[0]
                pvals = pvals.applymap(lambda x: min(x,1))
            elif adjust_pvals == "BH":
                pvals  = apply_bh(pvals, a = pval_cutoff)
        best_match_stats = {}
        best_pval = pvals.min(axis=1)
        for subt in known_groups[cl].keys():
            w = len(known_groups[cl][subt])/N # weight
            subt_pval = pvals[subt]
            passed_pvals = subt_pval[subt_pval==best_pval]
            passed_pvals = subt_pval[subt_pval<pval_cutoff].index.values
            d = performance.loc[passed_pvals,subt].sort_values(ascending=False)
            if d.shape[0] == 0:
                best_match_stats[subt]= {"bm_id":np.nan,
                                         performance_measure:0,"weight":w,
                                         "adj_pval":np.nan,
                                         "is_enriched":np.nan,
                                         "samples":set([]),"n_samples":0}
            else:
                bm_j = d.values[0]
                d = d[d==bm_j]
                bm_id = sorted(pvals.loc[d.index.values,:].sort_values(by = subt,ascending=True).index.values)[0]
                bm_pval = pvals.loc[bm_id,subt]
                bm_j = performance.loc[bm_id,subt]
                bm_is_enrich = is_enriched.loc[bm_id,subt]
                bm_samples = sample_clusters.loc[bm_id,"samples"]
                best_match_stats[subt]= {"bm_id":bm_id,
                                         performance_measure:bm_j,"weight":w,
                                         "adj_pval":bm_pval,
                                         "is_enriched":bm_is_enrich,
                                        "samples":bm_samples,
                                        "n_samples":len(bm_samples)}
                
        best_match_stats = pd.DataFrame.from_dict(best_match_stats).T
        best_match_stats["classification"] = cl
        best_matches.append(best_match_stats)
        performances[cl]= sum(best_match_stats[performance_measure]*best_match_stats["weight"])
        
    performances = pd.Series(performances)
    best_matches = pd.concat(best_matches,axis=0)
    return performances, best_matches

def evaluate_overlaps_ARI(biclusters, known_groups, all_elements):
    # compute exact Fisher's p-values and Jaccard overlaps for samples
    pvals = {}
    is_enriched = {}
    ARI = {}
    N = len(all_elements)
    all_elements_list = sorted(all_elements)
    # sanity check - biclusters
    for i in biclusters.index.values:
        bic_members = biclusters.loc[i, "samples"]
        if not bic_members.intersection(all_elements) == bic_members:
            print("bicluster {} elements {} are not in 'all_elements'".format(i, " ".join(
                bic_members.difference(all_elements))), file=sys.stderr)
            bic_members = bic_members.intersection(all_elements)
    # sanity check and sorting
    group_names = list(known_groups.keys())
    sorted_group_names = [group_names[0]]  # group names ordered by group size
    for group in group_names:
        group_members = known_groups[group]
        if not group_members.intersection(all_elements) == group_members:
            print(group, "elements are not in 'all_elements'", file=sys.stderr)
            return

        if group != group_names[0]:
            for gn in range(len(sorted_group_names)):
                if len(group_members) < len(known_groups[sorted_group_names[gn]]):
                    sorted_group_names = sorted_group_names[:gn] + [group] + sorted_group_names[gn:]
                    break
                elif gn == len(sorted_group_names) - 1:
                    sorted_group_names = [group] + sorted_group_names
    # print(sorted_group_names)
    for group in sorted_group_names:
        group_members = known_groups[group]
        pvals[group] = {}
        is_enriched[group] = {}
        ARI[group] = {}
        # binary vector for target cluster
        group_binary = np.zeros(len(all_elements_list))
        for j in range(len(all_elements_list)):
            if all_elements_list[j] in group_members:
                group_binary[j] =1
        for i in biclusters.index.values:
            bic = biclusters.loc[i, :]
            bic_members = bic["samples"]
            bic_binary = np.zeros(len(all_elements_list))
            for j in range(len(all_elements_list)):
                if all_elements_list[j] in bic_members:
                    bic_binary[j] =1
            # calculate ARI for 2 binary vectors:
            ARI[group][i] = adjusted_rand_score(group_binary, bic_binary)
                  
            # Fisher's exact test        
            shared = len(bic_members.intersection(group_members))
            bic_only = len(bic_members.difference(group_members))
            group_only = len(group_members.difference(bic_members))
            union = shared + bic_only + group_only
            pval = pvalue(shared, bic_only, group_only, N - union)
            pvals[group][i] = pval.two_tail 
            is_enriched[group][i] = False
            if pval.right_tail < pval.left_tail:
                is_enriched[group][i] = True   

    pvals = pd.DataFrame.from_dict(pvals).loc[:, sorted_group_names]
    is_enriched = pd.DataFrame.from_dict(is_enriched).loc[:, sorted_group_names]
    ARI = pd.DataFrame.from_dict(ARI).loc[:, sorted_group_names]
    return pvals, is_enriched, ARI

def apply_bh(df_pval, a = 0.05):
    # applies BH procedure to each column of p-value table
    df_adj = {}
    for group in df_pval.columns.values:
        bh_res, adj_pval = fdrcorrection(df_pval[group].fillna(1).values, alpha=a)
        df_adj[group] = adj_pval
    df_adj = pd.DataFrame.from_dict(df_adj)
    df_adj.index = df_pval.index
    return df_adj


def evaluate_overlaps(biclusters, known_groups, all_elements, dimension="samples"):
    # compute exact Fisher's p-values and Jaccard overlaps for samples
    pvals = {}
    is_enriched = {}
    jaccards = {}
    N = len(all_elements)
    # sanity check - biclusters
    for i in biclusters.index.values:
        bic_members = biclusters.loc[i, dimension]
        if not bic_members.intersection(all_elements) == bic_members:
            print("bicluster {} elements {} are not in 'all_elements'".format(i, " ".join(
                bic_members.difference(all_elements))), file=sys.stderr)
            bic_members = bic_members.intersection(all_elements)
    # sanity check and sorting
    group_names = list(known_groups.keys())
    sorted_group_names = [group_names[0]]  # group names ordered by group size
    for group in group_names:
        group_members = known_groups[group]
        if not group_members.intersection(all_elements) == group_members:
            print(group, "elements are not in 'all_elements'", file=sys.stderr)
            return

        if group != group_names[0]:
            for gn in range(len(sorted_group_names)):
                if len(group_members) < len(known_groups[sorted_group_names[gn]]):
                    sorted_group_names = sorted_group_names[:gn] + [group] + sorted_group_names[gn:]
                    break
                elif gn == len(sorted_group_names) - 1:
                    sorted_group_names = [group] + sorted_group_names
    # print(sorted_group_names)
    for group in sorted_group_names:
        group_members = known_groups[group]
        pvals[group] = {}
        is_enriched[group] = {}
        jaccards[group] = {}
        for i in biclusters.index.values:
            bic = biclusters.loc[i, :]
            bic_members = bic[dimension]
            shared = len(bic_members.intersection(group_members))
            bic_only = len(bic_members.difference(group_members))
            group_only = len(group_members.difference(bic_members))
            union = shared + bic_only + group_only
            pval = pvalue(shared, bic_only, group_only, N - union)
            pvals[group][i] = 1 
            if pval.right_tail < pval.left_tail:
                pvals[group][i] = pval.right_tail
                is_enriched[group][i] = True
            # if under-representation, flip query set
            else:
                pvals[group][i] = pval.left_tail # save left-tail p-value and record that this is not enrichment
                is_enriched[group][i] = False
                bic_members = all_elements.difference(bic_members)
                shared = len(bic_members.intersection(group_members))
                union = len(bic_members.union(group_members))
            jaccards[group][i] = shared / union

        # print(group,jaccards[group])

    pvals = pd.DataFrame.from_dict(pvals).loc[:, sorted_group_names]
    is_enriched = pd.DataFrame.from_dict(is_enriched).loc[:, sorted_group_names]
    jaccards = pd.DataFrame.from_dict(jaccards).loc[:, sorted_group_names]
    return pvals, is_enriched, jaccards


def calc_overlap_pval(overlap, group1_only,group2_only,background, max_N=5000):
    # if sample size < max_N), use Fisher's exact
    # otherwise replacing exact Fisher's with chi2 
    if overlap+group1_only+group2_only+background<max_N:
        pval = pvalue(overlap, group1_only,group2_only,background).right_tail
    else:
        chi2, pval, dof, expected = chi2_contingency([[overlap, group1_only],[group2_only,background]])
    return pval

def find_best_matching_biclusters(bics1, bics2, sizes, by="genes", adj_pval_thr=0.05, min_g = 2):
    # takes two biluster dafaframes from read_bic_table
    # by = "genes" or "samples" or "both"
    # sizes - dimensions of input matrix (n_genes,n_samples)
    # finds best matches of bics1 biclusters among bics2 biclusters

    N_g, N_s = sizes
    n_bics1 = bics1.shape[0]
    n_bics2 = bics2.shape[0]
    
    best_matches = {}  # OrderedDict({})
    for row1 in bics1.iterrows():
        bic1 = row1[1]
        i1 = row1[0]
        
        best_matches[i1] = {}
        bm_J = 0
        bm_o = 0
        bm_adj_pval = 1
        bm_id = None

        for row2 in bics2.iterrows():
            bic2 = row2[1]
            i2 = row2[0]
            
            g1 = bic1["genes"]
            s1 = bic1["samples"]
            g2 = bic2["genes"]
            s2 = bic2["samples"]
            o_g = len(g1.intersection(g2))
            o_s = len(s1.intersection(s2))
            s1 = len(s1)
            s2 = len(s2)
            g1 = len(g1)
            g2 = len(g2)
            J = 0
            adj_pval = 1
            # if not by="samples", ignore overlaps with gene < min_g
            if (by!="samples" and o_g >= min_g) or by=="samples": 
                if by=="genes" or by == "both":
                    g2_ = g2 - o_g # genes exclusively in bicluster 2
                    g1_ = g1 - o_g
                    u_g = g1_ + g2_ + o_g
                    bg_g = N_g - u_g
                    J_g = o_g * 1.0 / u_g
                    if not by == "both":
                        pval_g = calc_overlap_pval(o_g, g1_, g2_, bg_g)
                elif by=="samples" or by == "both":
                    s2_ = s2 - o_s # samples exclusively in bicluster 2
                    s1_ = s1 - o_s
                    u_s = s1_ + s2_ + o_s
                    bg_s = N_s - u_s
                    pval_s = calc_overlap_pval(o_s, s1_, s2_, bg_s)
                    # if p-val is high but one of the biclusters is large,
                    # try flipping the largest bicluster if it is close to 50% of the cohort                              
                    if pval_s>adj_pval_thr and max(s1,s2) > 0.4 * N_s:  
                        if s1 > s2:  # flip s1
                            s1 = N_s-s1
                            u_s = bg_s + s2
                            o_s = s2_
                            s2_ = s2 - o_s
                            bg_s = s1_
                            s1_ = s1 - o_s
                        else:  # flip s2
                            s2 = N_s - s2
                            u_s = bg_s + s1
                            o_s = s1_
                            s1_ = s1 - o_s
                            bg_s = s2_
                            s2_ = s2 - o_s
                        assert bg_s == N_s - u_s, "i1=%s; i2=%s: bg=%s, N_s=%s, u_s=%s"%(i1,i2,bg_s,N_s,u_s)
                        assert u_s == o_s + s1_ + s2_, "i1=%s; i2=%s: u_s=%s, o_s=%s, s1_=%s, s2_=%s"%(i1,i2,u_s,o_s,s1_,s2_)
                        if not by == "both":
                            # compute p-value again
                            pval_s = calc_overlap_pval(o_s, s1_, s2_, bg_s)
                    #try:
                    J_s = o_s * 1.0 / u_s
                    #except:
                    #    print( "i1=%s; i2=%s: u_s=%s, o_s=%s, s1_=%s, s2_=%s, bg_s=%s"%(i1,i2,u_s,o_s,s1_,s2_,bg_s))
                    #    J_s = 0
               
                if by == "genes":
                    J = J_g
                    pval = pval_g
                    o = o_g
                elif by == "samples":
                    J= J_s
                    pval = pval_s
                    o = o_s
                else:
                    o = o_s*o_g # bicluster overlap
                    b1_ = s1*g1-o # exclusive bicluster 1 area
                    b2_ = s2*g2 -o
                    u = o + b1_ + b2_
                    bg = N_s*N_g - u
                    J = o*1.0/u
                    pval = calc_overlap_pval(o, b1_, b2_, bg)
                        
                adj_pval = pval * n_bics2* n_bics1
                if adj_pval < adj_pval_thr and J>0:
                    if J > bm_J or (J == bm_J and adj_pval < bm_adj_pval):
                        bm_J = J
                        bm_adj_pval = adj_pval
                        bm_id = i2
                        bm_o = o
        best_matches[i1]["bm_id"] = bm_id
        best_matches[i1]["J"] = bm_J
        best_matches[i1]["adj_pval"] = bm_adj_pval
        if "genes" in bics1.columns and  "genes" in bics2.columns:
            if bm_id:
                best_matches[i1]["shared_genes"] = bics2.loc[bm_id, "genes"].intersection(bics1.loc[i1, "genes"])
                best_matches[i1]["n_shared_genes"] = len(best_matches[i1]["shared_genes"])
                best_matches[i1]["bm_genes"] = bics2.loc[bm_id,"genes"]
                best_matches[i1]["bm_n_genes"] = bics2.loc[bm_id,"n_genes"]
            best_matches[i1]["genes"] = bics1.loc[i1,"genes"]
            best_matches[i1]["n_genes"] = bics1.loc[i1,"n_genes"]
        if "samples" in bics1.columns and "samples" in bics2.columns:
            best_matches[i1]["n_samples"] = bics1.loc[i1,"n_samples"]
            best_matches[i1]["samples"] = bics1.loc[i1,"samples"]
            if bm_id:
                best_matches[i1]["bm_n_samples"] = bics2.loc[bm_id,"n_samples"]
                best_matches[i1]["bm_samples"] = bics2.loc[bm_id,"samples"]
                best_matches[i1]["shared_samples"] = bics2.loc[bm_id, "samples"].intersection(bics1.loc[i1, "samples"])
                best_matches[i1]["n_shared_samples"] = len(best_matches[i1]["shared_samples"])
            
                                                  
    best_matches = pd.DataFrame.from_dict(best_matches).T
    return best_matches


def bic_survival(surv_anno,samples,event = "OS",surv_time = "",
                 lr = True, verbose = True):
    # surival annotation - annotation matrix with time,event, and covariates
    # samples - samples in a group, e.g. biclsuter samples
    # check  complete separation
    # if all events are either inside or outside sample group
    if not surv_time:
        surv_time= event+".time"
    surv_data = surv_anno.copy()
    surv_data = surv_data.dropna(axis=0)
    
    # check zero variance columns:
    v =  surv_data.var()
    for col in v.index:
        if v[col] == 0:
            if verbose:
                print(col,"with zero variance excluded",file =sys.stderr)
            surv_data = surv_data.drop(col,axis=1)
    
    surv_data.loc[:,"x"] = 0
    surv_data.loc[list(set(samples).intersection(set(surv_data.index.values))),"x"] = 1

    pval = np.nan
    hr, upper_95CI, lower_95CI =  np.nan,np.nan,np.nan
    results = {}
    
    events = surv_data[event].astype(bool)
    
    v1 = surv_data.loc[events, 'x'].var()
    v2 = surv_data.loc[~events, 'x'].var()
    
    v3 = surv_data.loc[surv_data["x"]==1, event].var()
    v4 = surv_data.loc[surv_data["x"]==0, event].var()
    
    if v1 ==0 or v2 ==0:
        if verbose:
            in_bic = surv_data.loc[surv_data["x"]==1,:].shape[0]
            in_bg = surv_data.loc[surv_data["x"]==0,:].shape[0]
            print("perfect separation for biclsuter of  %s/%s samples"%(in_bic,in_bg),
                 "variances: {:.2f} {:.2f}".format(v1, v2), file = sys.stderr)
    if v3 == 0:
        print("zero variance for events in group; all events are ",set(surv_data.loc[surv_data["x"]==1, event].values))
    if v4 == 0:
        print("zero variance for events in background; all events are ",set(surv_data.loc[surv_data["x"]==0, event].values))
    
    # check variance of covariates in event groups
    exclude_covars =[]
    for c in [x for x in surv_data.columns.values if not x in ["x",event, surv_time]]:
        if surv_data.loc[events, c].var()==0:
            exclude_covars.append(c)
            print("\t",c,"variance is 0 in event group",file = sys.stdout)
        if surv_data.loc[~events, c].var()==0:
            exclude_covars.append(c)
            print("\t",c,"variance is 0 in no-event group",file = sys.stdout)
    #if len(exclude_covars)>0:
    #    cols = surv_data.columns.values
    #    cols = [x for x in cols if not x in exclude_covars]
    #    surv_data = surv_data.loc[:,cols]
    
    else:
        try:
            cph = CoxPHFitter()
            res = cph.fit(surv_data, duration_col=surv_time, event_col= event, show_progress=False)
            res_table = res.summary
            res_table  = res_table#.sort_values("p")    
            pval = res_table.loc["x","p"]
            hr = res_table.loc["x","exp(coef)"]
            upper_95CI = res_table.loc["x","exp(coef) upper 95%"]
            lower_95CI = res_table.loc["x","exp(coef) lower 95%"]
        except:
            pass
    
    results = {"p_value":pval,"HR":hr,
                  "upper_95CI":upper_95CI,
                  "lower_95CI":lower_95CI}
    # Log-rank test
    if lr:
        bic = surv_data.loc[surv_data["x"]==1,:]
        bg = surv_data.loc[surv_data["x"]==0,:]
        
        lr_result = logrank_test(bic.loc[:,surv_time], bg.loc[:,surv_time],
                                 event_observed_A=bic.loc[:,event],
                                 event_observed_B=bg.loc[:,event])
        results["LogR_p_value"] = lr_result.p_value
    
    
    return results

def add_survival(biclusters, # dataframes with biclustes
                 sample_data, # sample annotation
                 event= "OS", surv_time = "", # event and time column names
                 covariates=[],
                 min_n_events=5,
                 verbose = True):
    # if too few events, add na columns
    if sample_data[event].sum() < min_n_events:
        df = biclusters.copy()
        for col in [".p_value",".p_value_BH",
                    ".HR",".upper_95CI",".lower_95CI",
                    ".LogR_p_value",".LogR_p_value_BH"]:
            df[event+col] = np.nan
        return df
    if not surv_time:
        surv_time= event+".time"
    surv_results = {}
    for bic in biclusters.iterrows():
        sample_set = bic[1]["samples"]
        surv_data = sample_data.loc[:,covariates +[event,surv_time]]
        
        surv_results[bic[0]] = bic_survival(surv_data,
                                   sample_set,
                                   event = event,
                                   surv_time = surv_time,
                                   verbose = verbose)
        if "pval" in surv_results[bic[0]].keys():
            if np.isnan(surv_results[bic[0]]["pval"]):
                print("failed to fit CPH model for %s ~ bicluster %s"%(event,bic[0]),
                  file = sys.stderr)
        
    surv_results = pd.DataFrame.from_dict(surv_results).T
    surv_results.columns = [event+"."+x for x in surv_results.columns]
    
    pvals = surv_results.loc[~surv_results[event+".p_value"].isna(),event+".p_value"].values
    bh_res, adj_pval = fdrcorrection(pvals, alpha=0.05)
    surv_results.loc[~surv_results[event+".p_value"].isna(),event+".p_value_BH"] = adj_pval
    
    pvals = surv_results.loc[~surv_results[event+".LogR_p_value"].isna(),event+".LogR_p_value"].values
    bh_res, adj_pval = fdrcorrection(pvals, alpha=0.05)
    surv_results.loc[~surv_results[event+".LogR_p_value"].isna(),event+".LogR_p_value_BH"] = adj_pval
    
    return pd.concat([biclusters, surv_results],axis=1)


def test_sample_overlap(row, sample_set, N):
    # usage: 
    # biclusters_df.apply(lambda row: test_sample_overlap(row, sample_set, N),axis=1)
    # N - total number of samples in dataset
    bic_samples = row["samples"]
    o = len(sample_set.intersection(bic_samples))
    bic_only = len(bic_samples)-o
    sample_set_only = len(sample_set)-o
    bg = N-o-bic_only-sample_set_only
    p= pvalue(o,bic_only,sample_set_only,bg ).right_tail
    #if p<0.001:
    #    print(p,(o,bic_only,sample_set_only,bg),row["genes"])
    return pd.Series({"pval":p,"counts":(o,bic_only,sample_set_only,bg)})

def add_sex(biclusters,males = [],females=[]):
    sample_sets = {}
    #if len(males)>0:
    sample_sets["male"] = set(males)
    #if len(females)>0:
    sample_sets["female"] = set(females)
    
    N = len(males)+len(females)
    dfs =[]
    for sex in sample_sets.keys():
        sample_set = sample_sets[sex]
        df = biclusters.apply(lambda row: test_sample_overlap(row, sample_set, N),axis=1)
        df.columns = [sex+"."+x for x in df.columns]
        bh_res, adj_pval = fdrcorrection(df[sex+".pval"].values, alpha=0.05)
        df[sex+".pval_BH"] =  adj_pval
        dfs.append(df)
    dfs = pd.concat(dfs,axis=1)
    dfs["sex.pval_BH"] = dfs.loc[:,["male.pval_BH","female.pval_BH"]].min(axis=1)     
    dfs["sex"] = ""
    try:
        dfs.loc[dfs["male.pval_BH"]<0.05,"sex"] = "male"
    except:
        pass
    try:
        dfs.loc[dfs["female.pval_BH"]<0.05,"sex"] = "female"
    except:
        pass
    return pd.concat([biclusters,dfs],axis=1)


# Usage: from unpast.utils.unpast_DE import run_de_for_unpast
# run_de_for_unpast(unpast_output_path, expression_matrix_path, counts = False, [adj_p_value_cut_off = 0.05, logFC_cut_off = 1, r_script_path = None, r_executable_path = None])

import pandas as pd
import os
import logging
import subprocess
import numpy as np

# Add all logging levels
logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")

####################################################################################################
# Constants
####################################################################################################
DELIMITER = "\t"

####################################################################################################
# Functions
####################################################################################################


def extract_samples_to_file(unpast_df: pd.DataFrame, samples_to_compare: str) -> None:
    """
    Extract samples from the unpast dataframe.
    """
    samples = unpast_df[["samples", "n_samples"]]
    samples.to_csv(samples_to_compare, sep=DELIMITER, index_label="")


def run_add_genes_script(
    samples_to_compare: str,
    expression_matrix_path: str,
    counts: bool = False,
    adj_p_value_cut_off: float = 0.05,
    logFC_cut_off: float = 1,
    num_genes_cut_off = float('inf'),
    r_script_path: str = None,
    r_executable_path: str = None,
) -> str:
    if not r_script_path:
        # Assume add_genes.R is in the same folder
        r_script_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "add_genes.R")

    if r_executable_path is None:
        r_executable_path = "Rscript"
    else:
        r_executable_path = os.path.join(r_executable_path, "Rscript")

    logging.debug(
        [
            r_executable_path,
            r_script_path,
            samples_to_compare,
            expression_matrix_path,
            str(counts).upper(),
            str(adj_p_value_cut_off),
            str(logFC_cut_off),
            str(num_genes_cut_off)
        ]
    )

    process = subprocess.Popen(
        [
            r_executable_path,
            r_script_path,
            samples_to_compare,
            expression_matrix_path,
            str(counts).upper(),
            str(adj_p_value_cut_off),
            str(logFC_cut_off),
            str(num_genes_cut_off)
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = process.communicate()

    # Decode the output and error messages
    samples_with_genes_path = stdout.decode("utf-8")

    if stdout:
        logging.debug("DE analysis temporary output was generated: %s", samples_with_genes_path)
    if stderr:
        error_message = stderr.decode("utf-8")
        logging.error("Error Output: %s", error_message)
        raise Exception(f"Error in subprocess: {error_message}")

    return samples_with_genes_path


def filter_de_genes(new_unpast_df: pd.DataFrame, columns_to_process: list) -> pd.DataFrame:
    """
    Filter de_genes_df to keep only the genes that are in the unpast_df.
    """
    # Loop through each column
    for col in columns_to_process:
        col_DE = col + "_DE"

        # Check for NaN in original column and replace the corresponding _DE column value with NaN
        new_unpast_df[col_DE] = np.where(new_unpast_df[col].isna(), np.nan, new_unpast_df[col_DE])

        # Split the string into a list for both columns
        new_unpast_df[col] = new_unpast_df[col].apply(lambda x: str(x).split() if pd.notna(x) else [])
        new_unpast_df[col_DE] = new_unpast_df[col_DE].apply(lambda x: str(x).split() if pd.notna(x) else [])

        # Keep in the _DE column only values that are in the corresponding original column
        new_unpast_df[col_DE] = new_unpast_df.apply(lambda x: [i for i in x[col_DE] if i in x[col]], axis=1)

        # Join the list back into a string
        new_unpast_df[col] = new_unpast_df[col].apply(lambda x: " ".join(x) if x else np.nan)
        new_unpast_df[col_DE] = new_unpast_df[col_DE].apply(lambda x: " ".join(x) if x else np.nan)
    return new_unpast_df


def add_columns_to_unpast_df(unpast_df: pd.DataFrame, de_genes_df: pd.DataFrame) -> pd.DataFrame:
    """
    Filter de_genes_df to keep only the genes that are in the unpast_df.
    Add four new columns to the original unpast table.
    """
    # add columns to unpast_df
    new_unpast_df = unpast_df.assign(
        n_genes_DE=de_genes_df["n_genes"],
        genes_DE=de_genes_df["genes"],
        genes_up_DE=de_genes_df["genes_up"],
        genes_down_DE=de_genes_df["genes_down"],
    )

    new_unpast_df = filter_de_genes(new_unpast_df, columns_to_process=["genes", "genes_down", "genes_up"])

    # update n_genes_DE column with the number of genes in the genes_DE column
    new_unpast_df["n_genes_DE"] = new_unpast_df["genes_DE"].apply(lambda x: len(x.split()) if pd.notna(x) else 0)

    return new_unpast_df


def read_dataframe_from_file(file_path: str) -> pd.DataFrame:
    if not os.path.isfile(file_path):
        logging.error("File does not exist: %s", file_path)
        raise FileNotFoundError(f"File does not exist: {file_path}")

    if os.stat(file_path).st_size == 0:
        logging.error("File is empty: %s", file_path)
        raise ValueError(f"File is empty: {file_path}")

    df = pd.read_csv(file_path, delimiter=DELIMITER, header=0, skiprows=1, index_col=0)

    if df.empty:
        logging.error("UnPaSt output is empty: %s", file_path)
        raise ValueError(f"Dataframe is empty: {file_path}")

    return df


def write_result(df: pd.DataFrame, input_file_path: str, output_file_path: str) -> None:
    # Checking if input file exists and if it's not empty
    if not os.path.isfile(input_file_path):
        logging.error("Input file does not exist: %s", input_file_path)
        raise FileNotFoundError(f"Input file does not exist: {input_file_path}")

    if os.stat(input_file_path).st_size == 0:
        logging.error("Input file is empty: %s", input_file_path)
        raise ValueError(f"Input file is empty: {input_file_path}")

    # Check if output directory is writable
    output_dir_name = os.path.dirname(output_file_path)
    if not os.access(output_dir_name, os.W_OK):
        logging.error("Output directory is not writable: %s", output_dir_name)
        raise PermissionError(f"Output directory is not writable: {output_dir_name}")

    with open(input_file_path, "r") as f_in, open(output_file_path, "w") as f_out:
        comment_line = f_in.readline()
        f_out.write(comment_line)
        df.to_csv(f_out, sep=DELIMITER)


def safe_remove(file_path: str) -> None:
    try:
        os.remove(file_path)
    except OSError as e:
        logging.error(f"Error: {file_path} : {e.strerror}")


def run_de_for_unpast(
    unpast_output_path: str,
    expression_matrix_path: str,
    counts: bool = False,
    adj_p_value_cut_off: float = 0.05,
    logFC_cut_off: float = 1,
    num_genes_cut_off: float = float('inf'),
    r_script_path: str = None,
    r_executable_path: str = None,
) -> None:
    """
    A function that runs differential expression analysis using the limma package in R for unpast biclasters.
    """
    # Read unpast output
    # Read file with rownames in the first columns, first line is the comment line, colnames in the second line
    logging.info("Reading UnPaSt biclusters output and extracting samples to compare")
    # Check file existance, emptiness, and read DataFrame
    unpast_df = read_dataframe_from_file(unpast_output_path)
    # get the samples to compare
    samples_to_compare = os.path.join(
        os.path.dirname(unpast_output_path),
        f"{os.path.splitext(os.path.basename(unpast_output_path))[0][:-len('_biclusters')]}_samples{os.path.splitext(os.path.basename(unpast_output_path))[1]}",
    )
    # save samples to file
    extract_samples_to_file(unpast_df, samples_to_compare)

    # run add_genes.R script
    logging.info("Running DE analysis for unpast biclasters, takes a while...")
    samples_with_genes_path = run_add_genes_script(
        samples_to_compare,
        expression_matrix_path,
        counts,
        adj_p_value_cut_off,
        logFC_cut_off,
        num_genes_cut_off,
        r_script_path,
        r_executable_path,
    )

    # Read the unpast output file into a pandas dataframe.
    # Read file with rownames in the first columns, first line is the comment line, colnames in the second line
    logging.info("Filtering DE genes and adding them to the original UnPaSt output")
    de_genes_df = pd.read_csv(samples_with_genes_path.strip(), delimiter=DELIMITER, header=0, index_col=0)

    # filter de_genes_df to keep only the genes that are in the unpast_df
    new_unpast_df = add_columns_to_unpast_df(unpast_df, de_genes_df)

    # write new_unpast_df to file
    # use the original unpast_df file name and add _DE to the end
    # also add the comment line from the original unpast_df to the top of the new file
    output_path_de = unpast_output_path.replace(".tsv", f'_DE.pval{adj_p_value_cut_off}.logFC{logFC_cut_off}.tsv')
    write_result(new_unpast_df, unpast_output_path, output_path_de)

    # remove the temporary files
    logging.info("Removing temporary files")
    safe_remove(samples_to_compare)
    safe_remove(samples_with_genes_path.strip())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['unpast', 'unpast.utils']

package_data = \
{'': ['*']}

install_requires = \
['fisher==0.1.10',
 'jenkspy==0.2.0',
 'kneed>=0.8.3,<0.9.0',
 'lifelines==0.27.8',
 'matplotlib-venn==0.11.6',
 'numba==0.55.2',
 'numpy==1.22.3',
 'pandas==1.4.2',
 'python-louvain==0.15',
 'scikit-learn==1.1.0',
 'scikit-network==0.25.0',
 'scipy==1.7.3',
 'statsmodels==0.13.2']

entry_points = \
{'console_scripts': ['run_unpast = unpast.run_unpast:main']}

setup_kwargs = {
    'name': 'unpast',
    'version': '0.1.9.4',
    'description': 'A novel method for unsupervised patient stratification.',
    'long_description': '# UnPaSt\n\nUnPaSt is a novel method for identification of differentially expressed biclusters.\n\n![alt text](./poster/DESMOND2_steps2.png)\n\n\n## Requirements:\n<pre>\nPython:\n    fisher==0.1.10\n    jenkspy==0.2.0\n    matplotlib-venn==0.11.6\n    numba==0.55.2\n    numpy==1.22.3\n    scikit-learn==1.1.0\n    scikit-network==0.25.0\n    scipy==1.7.3\n    statsmodels==0.13.2\n    pandas==1.4.2\n    python-louvain==0.15\n    statsmodels==0.13.2\n\nR:\n    WGCNA>=1.70-3\n    limma>=3.42.2\n</pre>\n\n\n## Installation\n* UnPaSt can be installed using `pip`, `poetry`, or run using `Docker`, or as a script (see examples section). Follow the appropriate instructions below for your preferred method. You need to have R and **Python 3.8-3.10** installed.\n\n1. Using **pip**: \\\n    To install the project using `pip`, first make sure you have `pip` installed on your system. If you haven\'t installed it already, you can find the installation instructions [here](https://pip.pypa.io/en/stable/installation/). \\\n    Once `pip` is installed, you can install UnPaSt by running the following command:\n\n    ```bash\n    pip install unpast\n    ```\n    Run it:\n    ```bash\n    run_unpast -h\n    ```\n    **Dependencies**. To use this package, you will need to have R and the [WGCNA library](https://horvath.genetics.ucla.edu/html/CoexpressionNetwork/Rpackages/WGCNA/) and [limma](https://bioconductor.org/packages/release/bioc/html/limma.html) installed. You can easily install these dependencies by running the following command after installing unpast:\n    ```bash\n    python -m unpast.install_r_dependencies\n\n    # or you can install it directly\n    R -e "install.packages(\'BiocManager\'); BiocManager::install(c(\'WGCNA\', \'limma\'))"\n    ```\n\n2. Installation using **Poetry**: \\\n    To install the package using Poetry, first make sure you have Poetry installed, clone the repo and run:\n    ```bash\n    poetry add unpast\n    ```\n    Run it:\n    ```bash\n    poetry run run_unpast -h\n    ```\n    **Dependencies**. To use this package, you will need to have R and the [WGCNA library](https://horvath.genetics.ucla.edu/html/CoexpressionNetwork/Rpackages/WGCNA/) and [limma](https://bioconductor.org/packages/release/bioc/html/limma.html) installed. You can easily install these dependencies by running the following command after installing unpast:\n    ```bash\n    poetry run python -m unpast.install_r_dependencies\n\n    # or you can install it directly\n    R -e "install.packages(\'BiocManager\'); BiocManager::install(c(\'WGCNA\', \'limma\'))"\n    ```\n3. Running with **Docker**: \\\n    You can also run the package using Docker. First, pull the Docker image:\n    ```bash\n    docker pull freddsle/unpast:latest\n    ```\n    Next, run the UnPaSt:\n    ```bash\n    docker run -v /your/data/path/:/user_data/ freddsle/unpast:latest --exprs /user_data/exprs.tsv --out_dir /user_data/out_dir/\n    ```\n\n\n## Examples\n* UnPaSt requires a tab-separated file with features (e.g. genes) in rows, and samples in columns. Feature and sample names must be unique. \n\n<pre>\n\ncd test;\nmkdir -p results;\n\n# running UnPaSt with default parameters and example data\npython ../run_unpast.py --exprs scenario_B500.exprs.tsv.gz --basename results/scenario_B500\n\n# with different binarization and clustering methods\npython ../run_unpast.py --exprs scenario_B500.exprs.tsv.gz --basename results/scenario_B500 --binarization ward --clustering Louvain\n\n# help\npython run_unpast.py -h\n</pre>\n\n## Outputs\n* \\<basename\\>.[parameters].biclusters.tsv - a .tsv table with found biclsuters, where \n    - the first line starts from \'#\' and stores parameters\n    - each following line represents a bicluster\n    - SNR column contains SNR of a bicluster \n    - columns "n_genes" and "n_samples" provide the numbers of genes and samples, respectively \n    - "gene","sample" contain gene and sample names respectively\n    - "gene_indexes" and  "sample_indexes" - 0-based gene and sample indexes in the input matrix.\n* binarized expressions, background distributions of SNR for each bicluster size and binarization statistics [if clustering is WGCNA,  or  \'--save_binary\' flag is added]\n\n## About \nUnPaSt is an unconstrained version of DESMOND method ([repository](https://github.com/ozolotareva/DESMOND), [publication](https://academic.oup.com/bioinformatics/article/37/12/1691/6039116?login=true))\n\nMajor modifications:\n * it does not require the network of feature interactions \n * UnPaSt clusters individual features instead of pairs of features\n * uses 2-means, hierarchicla clustering or GMM for binarization of individual gene expressions\n * SNR threshold for featuer selection is authomatically determined; it depends on bicluster size in samples and user-defined p-value cutoff\n \n## License\nFree for non-for-profit use. For commercial use please contact the developers. \n',
    'author': 'Olga Zolotareva (ozolotareva)',
    'author_email': 'None',
    'maintainer': 'Olga Zolotareva (ozolotareva)',
    'maintainer_email': 'None',
    'url': 'https://github.com/ozolotareva/DESMOND2',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import sys
import copy
import random
import pandas as pd
import numpy as np
from time import time
import math
import os

import matplotlib.pyplot as plt

# optimizer
TRY_USE_NUMBA=True
def jit_if_available(func):
    # default "do nothing" decorator with the numba-like interface
    def decorated(*args, **kwargs):
        return func(*args, **kwargs)
    return decorated
if TRY_USE_NUMBA:
    try:
        from numba import jit  # as jit_if_available
        jit_if_available = jit()
    except:
        print("Numba is not available. Install numba for a bit faster calculations")


################## 2. Probabiliatic clustering #############
@jit_if_available
def calc_lp(gene_ndx,module_ndx,gene2Samples,
            nOnesPerSampleInModules,moduleSizes,
            moduleOneFreqs,p0,match_score,mismatch_score,bK_1,N,
            alpha,beta_K):
    
    # 1. Prepare vals: (n_ones_per_pat, m_size, alpha, gene_vector, beta_K)
    #      and return answer in special cases
    
    m_size = moduleSizes[module_ndx]
    if m_size == 0:
        # gene is removed, the module is empty
        return p0
    
    gene_vector = gene2Samples[gene_ndx,] 
    n_ones_per_pat = nOnesPerSampleInModules[module_ndx,]
    if gene_ndx == module_ndx: # remove the gene from its own module
        if m_size == 1:
            return p0
        m_size -=1
        n_ones_per_pat = n_ones_per_pat - gene_vector

    # if a module is composed of a single gene
    if m_size == 1:
        # just count number of matches and mismatches and
        # n_matches =  np.inner(n_ones_per_pat,gene_vector)
        n_matches =  np.sum(n_ones_per_pat[gene_vector==1])

        return n_matches*match_score + (N-n_matches)*mismatch_score + bK_1
    
    # 2. usual case if a module contains more than one gene
    return calc_lp_formula(n_ones_per_pat, m_size, alpha, gene_vector, beta_K)

@jit_if_available
def calc_lp_formula(n_ones_per_pat, m_size, alpha, gene_vector, beta_K): 
    beta_term = math.log(m_size+beta_K)
    
    # alpha_term
    # ones-matching
    oneRatios = (n_ones_per_pat+alpha/2)/(m_size+alpha)
    ones_matching_term = np.sum(np.log(oneRatios)[gene_vector == 1])

    # zero-matching
    # zeroRatios = (m_size-n_ones_per_pat+alpha/2)/(m_size+alpha)
    zeroRatios = 1 - oneRatios
    zeros_matching_term = np.sum(np.log(zeroRatios)[gene_vector == 0])

    return ones_matching_term + zeros_matching_term + beta_term

def calc_lp_column(module_ndx,gene2Samples,
            nOnesPerSampleInModules,moduleSizes,
            moduleOneFreqs,p0,match_score,mismatch_score,bK_1,N,
            alpha,beta_K):
    """Same as calc_lp, but for all the genes simultaneously"""
    m_size = moduleSizes[module_ndx]
    if m_size == 0:
        # all genes are removed, the module is empty, alfl answers are p0
        return p0

    n_ones_per_pat = nOnesPerSampleInModules[module_ndx,]
    if m_size == 1:
        # if a module is composed of a single gene
        n_matches = np.dot(gene2Samples, n_ones_per_pat)
        vals = n_matches*match_score + (N-n_matches)*mismatch_score + bK_1

    else: 
        # ones-matching
        oneRatios = (n_ones_per_pat+alpha/2)/(m_size+alpha)
        # ones_matching_term = np.dot((gene2Samples == 1), np.log(oneRatios))
        ones_matching_term = np_faster_dot(gene2Samples == 1, np.log(oneRatios))

        # zero-matching
        zeroRatios = 1 - oneRatios  # = (m_size-n_ones_per_pat+alpha/2)/(m_size+alpha)
        # zeros_matching_term = np.dot((gene2Samples == 0), np.log(zeroRatios))
        zeros_matching_term = np_faster_dot(gene2Samples == 0, np.log(zeroRatios))

        beta_term = math.log(m_size+beta_K)
        vals = ones_matching_term + zeros_matching_term + beta_term

    # calc LP[m,m] with the less optimized func
    vals[module_ndx] = calc_lp(module_ndx,module_ndx,gene2Samples,nOnesPerSampleInModules,moduleSizes,
                            moduleOneFreqs,p0,match_score,mismatch_score,bK_1,N,alpha,beta_K)
    return vals

@jit_if_available
def np_faster_dot(np_a, np_b): 
    # translating to float64 for jit compilation
    # may be a bit faster in some cases
    return np.dot(
        np_a.astype(np.float64),
        np_b.astype(np.float64)
    )

def set_initial_conditions(df, alpha,beta_K,verbose = True):
    t_0 = time()
    N = df.shape[0] # number of samples
    K = df.shape[1] # initial number of modules
    p0 = N*np.log(0.5)+np.log(beta_K)
    match_score = np.log((alpha*0.5+1)/(alpha))
    mismatch_score = np.log((alpha*0.5+0)/alpha)
    bK_1 = math.log(1+beta_K)
    #print("\t\tKxN=%sx%s"%(K,N))
    #print("\t\tp0=",p0)
    
    # p0, match_score, mismatch_score, bK_1
    genes = df.columns.values
    # 1. the number of genes inside each component, initially 1 for each gene
    moduleSizes=np.ones(K,dtype=np.int)
    
    # 2. a binary (int) matrix of size KxN that indicates the samples genes
    gene2Samples = df.T.values
    
    # 3. a binary matrix of size K by m that stores the total number of ones per sample in each module,
    # initially equal to 'gene2Samples'
    nOnesPerSampleInModules = copy.copy(gene2Samples)

    #4. initial module id
    gene2Module = list(range(0,K))
    gene2Module = np.array(gene2Module)
    
    #5. moduleOneFreqs
    moduleOneFreqs = []
    for g in range(0,K):
        moduleOneFreqs.append(float(sum(gene2Samples[g,]))/N)
    moduleOneFreqs = np.array(moduleOneFreqs)
    
    #6. setting initial LPs: i = gene, j=module
    LP = np.zeros((K,K),dtype=np.float)
    for i in range(0,K):
        #if (i+1)% 1000==0:
        #    if verbose:
        #        print("\t",i+1,"genes processed in ",round(time()- t_0,1),"s")
        for j in range(i,K):
            LP[i,j] = calc_lp(i,j,gene2Samples,
            nOnesPerSampleInModules,moduleSizes,
            moduleOneFreqs,p0,match_score,mismatch_score,bK_1,N,alpha, beta_K)
            LP[j,i] = LP[i,j]
    print("time:\tInitial state created in",round(time()-t_0, 1) , "s.", file = sys.stdout)
    return moduleSizes, gene2Samples, nOnesPerSampleInModules, gene2Module, moduleOneFreqs, LP

def adjust_lp(log_probs,n_exp_orders=7):
    # adjusting the log values before normalization to avoid under-flow
    max_p = max(log_probs)
    probs = []
    for lp in log_probs:
        # shift all probs to set max_prob less than log(max_np.float)  
        adj_lp = lp - max_p
        # set to minimal values all which less then 'n_orders' lesser than p_max to zeroes
        if adj_lp >= - n_exp_orders:
            probs.append(np.exp(adj_lp))
        else:
            probs.append(0)
    probs = probs/sum(probs)
    return probs

### functions for checking of convergence conditions ###

def count_skipping_genes(labels):
    n_skipping_genes = 0
    for gene in range(0,labels.shape[1]):
        states = labels[:,gene]
        unique = np.unique(states)
        if len(unique)> 1:
            n_skipping_genes +=1
    return n_skipping_genes


def check_convergence_conditions(n_skipping,n_min,n_max,tol=0.05, verbose = True):
    n_points = len(n_skipping)
    n_skipping = np.array(n_skipping,dtype=float)
    
    if n_max - n_min >0:
        # scale
        n_skipping = (n_skipping-n_min)/(n_max - n_min)*n_points
        # fit line
        A = np.vstack([range(0,n_points), np.ones(n_points)]).T
        k,b = np.linalg.lstsq(A, n_skipping, rcond=None)[0] # y = kx + b
    else:
        k = 0 
    
    if abs(k)<tol:
        convergence = True
    else:
        convergence = False
    if verbose:
        print("\t\t\tconvergence condition:",convergence,"; curve slope: {:.2f}".format(k), file=sys.stdout)
    return convergence  

# @jit_if_available
def apply_changes(gene_ndx,new_module,curr_module,LP,gene2Samples, gene2Module, nOnesPerPatientInModules,moduleSizes,
                moduleOneFreqs, p0,match_score,mismatch_score, bK_1,alpha,beta_K,N,K, calc_LPs=True):
    """Moves the gene from the current module to the new one
        and updates gene2Module, nOnesPerPatientInModules and moduleSizes respectively.
        K - quantity of genes
        M (=K) - quantitiy of modules
        S - quantity of samples (patients)
        
        see also set_initial_conditions for more info

    Args: 
        gene_ndx(int): 
        new_module(int): 
        curr_module(int):
        LP(np.array KxM): matrix of transmission probabilities
        gene2Samples(np.array KxS): matrix of samples co-regulated with genes
        gene2Module(list): list of i'th module
        nOnesPerPatientInModules(np.array MxS): precalculated sums for optimization
            nOnesPerPatientInModules[m] = sum(gene2Samples[g] for g in genes if gene2Module[g] == m)
        moduleSizes(np.array M) - genes in module
        
        moduleOneFreqs - is not used
        N - is not used

        p0(float) - precalculated default probability
        match_score(float) - precalculated const
        mismatch_score(float) - precalculated const
        bK_1(float) - precalculated const
        
        alpha(float) - alg. global parameter
        beta_K(float) - alg. global parameter

        K(int) - count of Genes

        calc_LPs(bool) - should LPs be recalculated? 

    """
    # update the gene module membership
    gene2Module[gene_ndx] = new_module
    # for this edge no probabilities change
    
    # reduce curr_module size and nOnesPerPatientInModules
    gene_vector = gene2Samples[gene_ndx,]
    nOnesPerPatientInModules[curr_module,] = nOnesPerPatientInModules[curr_module,] - gene_vector
    moduleSizes[curr_module,]-=1
    
    # increase new_module
    nOnesPerPatientInModules[new_module,] = nOnesPerPatientInModules[new_module,] + gene_vector
    moduleSizes[new_module,]+=1
    
    # update LPs for all genes contacting curr and new modules
    if calc_LPs:
        for module in curr_module, new_module:
            ndx_old_val = LP[gene_ndx, module] # for this gene no probabilities changed
            LP[:, module] = calc_lp_column(module,gene2Samples,nOnesPerPatientInModules,moduleSizes,
                                                            moduleOneFreqs,p0,match_score,mismatch_score,bK_1,N,alpha,beta_K)
            LP[gene_ndx, module] = ndx_old_val






def sampling(LP,gene2Module, gene2Samples,nOnesPerPatientInModules,moduleSizes,
             moduleOneFreqs, p0, match_score, mismatch_score, bK_1, alpha, beta_K,
             max_n_steps=100,n_steps_averaged = 5,n_points_fit = 10,tol = 0.05,
             n_steps_for_convergence = 10,verbose=True):
    
    K = len(gene2Module)
    N  = gene2Samples.shape[1]
    t_ =  time()
    gene2Module_history = [copy.copy(gene2Module)]
    sampling_steps = n_steps_for_convergence 
    is_converged = False
    
    for step in range(1, max_n_steps):
        #if verbose:
        #    print("step", step,file = sys.stdout)
        not_changed_genes = 0
        t_0 = time()
        t_1=t_0
        i = 1
        gene_ndxs = np.arange(K)
        np.random.shuffle(gene_ndxs)
        for gene_ndx in gene_ndxs:
            # adjust LogP and sample a new module
            P_adj = adjust_lp(LP[gene_ndx,:], n_exp_orders=7)
            curr_module = gene2Module[gene_ndx]
            new_module = np.random.choice(range(0,K), p=P_adj) 

            # update network and matrices if necessary
            if new_module != curr_module:
                apply_changes(gene_ndx,new_module,curr_module,LP,gene2Samples, gene2Module, nOnesPerPatientInModules,moduleSizes,
                moduleOneFreqs, p0,match_score,mismatch_score, bK_1,alpha,beta_K, N, K)
                
            else:
                not_changed_genes +=1#
            i+=1
            #if i%1000 == 0:
            #    if verbose:
            #        print(i,"\t\tgenes processed in",round(time()- t_1, 1) , "s runtime...",file=sys.stdout)
            #    t_1 = time()
        if verbose:
            print("\t\tstep ",step,round(time() - t_0, 1) , "s", file = sys.stdout)
        
        gene2Module_history.append(copy.copy(gene2Module))
        
        if step == n_steps_averaged:
            is_converged = False
            n_times_cc_fulfilled = 0
            labels = np.asarray(gene2Module_history[step-n_steps_averaged:step])
            n_skipping = [count_skipping_genes(labels)]
            
        if step > n_steps_averaged:
            labels = np.asarray(gene2Module_history[step-n_steps_averaged:step])
            n_skipping.append(count_skipping_genes(labels))
            if verbose:
                print("\t\t\tfeatures oscilating in the last {} steps: {}".format(n_steps_averaged, n_skipping[-1]),file=sys.stdout)
            
        
        if  step >= n_steps_averaged + n_points_fit:
            last_steps = n_skipping[-n_points_fit:]
            if len(np.unique(last_steps))==1: # stop when n_skipped is the same in last n_points_fit steps 
                if verbose:
                    print("The model converged after", step+1,"steps.", file = sys.stdout)
                    print("Consensus of last",sampling_steps,"states will be taken")
                    print("Sampling runtime",round(time()- t_ ,1) , "s", file = sys.stdout)
                return gene2Module_history, sampling_steps,n_skipping
            
            n_min, n_max = min(n_skipping),max(n_skipping)
            is_converged = check_convergence_conditions(last_steps,n_min, n_max,tol=tol,verbose = verbose)
                
            if is_converged:
                n_times_cc_fulfilled +=1
            else:
                n_times_cc_fulfilled = 0

            if n_times_cc_fulfilled > n_steps_for_convergence: # stop if convergence is True for the last n steps
                if verbose:
                    print("The model converged after", step+1,"steps.", file = sys.stdout)
                    print("Consensus of last",sampling_steps,"states will be taken")
                    print("Sampling runtime",round(time()- t_ ,1) , "s", file = sys.stdout)
                return gene2Module_history, sampling_steps,n_skipping
    
    if verbose:
        print("The model did NOT converge after", step+1,"steps.", file = sys.stdout)
        print("Consensus of last",sampling_steps,"states will be taken")
        print("Sampling runtime",round(time()- t_ ,1) , "s", file = sys.stdout)
        
    return gene2Module_history,sampling_steps,n_skipping


def plot_convergence(n_skipping_edges,thr_step,
                     alpha=1.0, beta_K=1.0,
                     n_steps_averaged=10,n_points_fit = 20, 
                     n_steps_for_convergence = 10,
                     save_plot = False):
    
    # plots numnber of oscilating genes
    steps = range(n_steps_averaged,n_steps_averaged+len(n_skipping_edges))
    fig, ax = plt.subplots(1, 1,figsize=(15,5))
    ax.set_title("Model convergence")
    ax.plot(steps, n_skipping_edges,'b.-')
    ax.axvline(thr_step,color="red",linestyle='--') 
    ax.set_ylabel("#genes oscilating on the last "+str(int(n_steps_averaged))+" steps")
    ax.set_xlabel('step')
    text = 'alpha=%s; beta/K=%s;\n'%(alpha,beta_K)
    text += "ns_a=%s; n_points_fit=%s; ns_c=%s"%(n_steps_averaged,n_points_fit,n_steps_for_convergence)
    ax.text(0.5, 0.90, text,
    verticalalignment='top', horizontalalignment='center',
    transform=ax.transAxes,
    color='black', fontsize=15)
    
    if save_plot:
        plt.savefig(save_plot, transparent=True)  


def get_consensus_modules(gene2module_history, f = 0.25, verbose = False):
    
    # g->m states from sampling phase
    labels = np.asarray(gene2module_history)
    
    # identify modules which genes ocsilate
    K = labels.shape[1] # number of genes and modules
    n_steps = labels.shape[0] # number of steps in sampling phase
    
    # consensus module -> genes 
    consensus = [[] for i in range(0,K)]
    
    for gene_ndx in range(0,K):
        unique, counts = np.unique(labels[:,gene_ndx], return_counts=True)
        
        if len(unique) == 1:
            consensus[unique[0]].append(gene_ndx) 
            #print("Gene %s -> [%s]" %(gene_ndx, ",".join(map(str,unique))),file= sys.stdout)
        else:
            freqs = np.array(counts)/n_steps
            new_ndxs  = [unique[i] for i in range(len(freqs)) if freqs[i]>=f]
            for i in new_ndxs:
                consensus[i].append(gene_ndx) 
            consensus.append(new_ndxs)
            
    if verbose:
        print()
        print("size(genes)\tn_modules")
        module_sizes = {}
        for m in consensus:
            s = len(m)
            if s in module_sizes.keys():
                module_sizes[s] +=1
            else:
                module_sizes[s] =1
        for s in sorted(module_sizes.keys()):
            print(s, "\t\t", module_sizes[s])
            
    return consensus



def run_sampling(exprs_bin,alpha=1.0,beta_K=1.0,f=0.33,
                 max_n_steps=50, n_steps_averaged = 10,
                 n_points_fit = 10, tol = 0.1, n_steps_for_convergence = 10,
                 verbose =True,plot_all=True):

    moduleSizes, gene2Samples, nOnesPerSampleInModules, gene2Module, moduleOneFreqs, LP  = set_initial_conditions(exprs_bin,alpha,beta_K,verbose = verbose)
    K = len(moduleSizes)
    N = gene2Samples.shape[1]
    if verbose:
        print("\t\tLP memory, {:.2f} M".format(LP.nbytes/(1024*1024)),file = sys.stdout)
    
    # simplifying probability calculations
    max_log_float = np.log(np.finfo(np.float64).max)
    n_exp_orders = 7 # ~1000 times 
    p0 = N*np.log(0.5)+np.log(beta_K)
    match_score = np.log((alpha*0.5+1)/(alpha))
    mismatch_score = np.log((alpha*0.5+0)/alpha)
    bK_1 = math.log(1+beta_K)
    genes = exprs_bin.columns.values
    
    t0 = time()
    gene2Module_history,sampling_steps,n_skipping_genes = sampling(LP,gene2Module, gene2Samples, nOnesPerSampleInModules,moduleSizes,moduleOneFreqs,
                                                                   p0, match_score,mismatch_score, bK_1, alpha, beta_K, 
                                                                   max_n_steps=max_n_steps, n_steps_averaged = n_steps_averaged, 
                                                                   n_points_fit = n_points_fit, tol = tol, 
                                                                   n_steps_for_convergence = n_steps_for_convergence, verbose=verbose)

    print("time:\tSampling ({} steps) fininshed in {:.2f} s.".format(len(gene2Module_history),time()-t0), file = sys.stdout)

    # plot the number of oscilating genes per step
    if plot_all:
        #suffix = ".alpha="+str(alpha)+",beta_K="+str(beta_K)+",pv="+str(snr_pval)
        #save_plot = out_dir + basename +suffix + ".convergence.svg"
        save_plot = False
        plot_convergence(n_skipping_genes[0:], len(gene2Module_history)-sampling_steps-0,
                                 alpha=alpha, beta_K=beta_K,
                                 n_steps_averaged=n_steps_averaged,
                                 n_points_fit = n_points_fit, 
                                 n_steps_for_convergence = n_steps_for_convergence,
                                 save_plot=save_plot)
    
    # take the last (n_points_fit+n_steps_for_convergence) steps modules:
    # and get consensus feature-to-module membership
    consensus = get_consensus_modules(gene2Module_history[-sampling_steps:],verbose= verbose,f = f)
    
    # 
    modules = []
    not_clustered = []
    
    ints2g_names = exprs_bin.columns.values
    for mid in range(0,len(consensus)):
        gene_ids = consensus[mid]
        if len(gene_ids) > 1:
            modules.append(ints2g_names[gene_ids].tolist())
        elif len(gene_ids) == 1:
            not_clustered.append(ints2g_names[gene_ids[0]])
            
    if verbose:
        print("\t{} modules and {} not clustered genes".format(len(modules),len(not_clustered)),file = sys.stdout)
        print("time\tmodules detected in in {:.2f} s.".format(time()-t0),file = sys.stdout)
    return (modules,not_clustered)
